package be.cylab.java.wowa.training;

import org.datavec.api.records.reader.RecordReader;
import org.datavec.api.records.reader.impl.csv.CSVRecordReader;
import org.datavec.api.split.FileSplit;
import org.datavec.api.util.ClassPathResource;
import org.deeplearning4j.datasets.datavec.RecordReaderDataSetIterator;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.eval.ROC;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.SplitTestAndTrain;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.preprocessor.DataNormalization;
import org.nd4j.linalg.dataset.api.preprocessor.NormalizerStandardize;
import org.nd4j.linalg.lossfunctions.LossFunctions;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;

/**
 * Class to test neural network learning.
 */

public final class NeuralNetwork {

    /**
     * Class count.
     */
    public static final int CLASSES_COUNT = 2;
    /**
     * Features count.
     */
    public static final int FEATURES_COUNT = 5;


    private HyperParameters parameters;


    /**
     * Default constructor.
     *
     * @param parameters
     */
    public NeuralNetwork(final HyperParameters parameters
    ) {

        this.parameters = parameters;
    }

    /**
     * @param learning
     * @return
     */
    MultiLayerNetwork learning(
            final DataSet learning) {

        int neurons_number = parameters.getNeuronsNumber();
        MultiLayerConfiguration configuration
                = new NeuralNetConfiguration.Builder()
                .iterations(1000)
                .activation(parameters.getActivationFunction())
                .optimizationAlgo(
                        parameters.getAlgorithm())
                .weightInit(WeightInit.XAVIER)
                .learningRate(parameters.getLearningRate())
                .regularization(true).l2(0.0001)
                .list()
                .layer(0, new DenseLayer.Builder().nIn(FEATURES_COUNT)
                        .nOut(neurons_number).build())
                .layer(1, new DenseLayer.Builder().nIn(neurons_number)
                        .nOut(neurons_number).build())
                .layer(2, new OutputLayer.Builder(LossFunctions
                        .LossFunction.NEGATIVELOGLIKELIHOOD)
                        .activation(Activation.SOFTMAX)
                        .nIn(neurons_number)
                        .nOut(CLASSES_COUNT).build())
                .backprop(true).pretrain(false)
                .build();

        MultiLayerNetwork model = new MultiLayerNetwork(configuration);
        model.init();
        model.fit(learning);

        return model;
    }

    /**
     * @param data
     * @param expected
     * @return
     */
    public MultiLayerNetwork run(
            final List<List<Double>> data,
            final List<Double> expected) {

        DataSet all_data = Utils.createDataSet(data, expected);
        DataSet training_data
                = prepareDataSetForTrainingAndTesting(all_data).getTrain();
        return this.learning(training_data);

    }

    /**
     * @param data_filename
     * @param expected_filename
     * @return
     */
    public MultiLayerNetwork run(
            final String data_filename,
            final String expected_filename) {
        List<List<Double>> data
                = Utils.convertJsonToDataForTrainer(data_filename);
        List<Double> expected
                = Utils.convertJsonToExpectedForTrainer(expected_filename);
        return this.run(data, expected);
    }

    /**
     * @param filename
     * @return
     */
    public MultiLayerNetwork runCSV(final String filename) {
        try (RecordReader record_reader = new CSVRecordReader(0, ',')) {
            record_reader.initialize(new FileSplit(
                    new ClassPathResource(filename).getFile()
            ));
            DataSetIterator iterator = new RecordReaderDataSetIterator(
                    record_reader, 12468, FEATURES_COUNT, CLASSES_COUNT);
            DataSet all_data = iterator.next();
            DataSet training_data
                    = prepareDataSetForTrainingAndTesting(all_data).getTrain();
            return learning(training_data);

        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Method to evaluate the performance of the model.
     *
     * @param testing
     * @param network
     */
    public Double modelEvaluation(
            final DataSet testing,
            final MultiLayerNetwork network) {
        INDArray output
                = network.output((testing.getFeatureMatrix()));

        Evaluation eval = new Evaluation(CLASSES_COUNT);
        eval.eval(testing.getLabels(), output);
        System.out.println(eval.stats());
        ROC roc = new ROC(CLASSES_COUNT);
        roc.eval(testing.getLabels(), output);
        System.out.println(roc.stats());
        return roc.calculateAUC();
    }

    public Double modelEvaluation(
            final TrainingDataset testing,
            final MultiLayerNetwork network) {
        DataSet test
                = Utils.createDataSet(testing.getData(), testing.getExpected());
        return modelEvaluation(test, network);
    }

    /**
     * @param data
     * @return
     */
    SplitTestAndTrain prepareDataSetForTrainingAndTesting(final DataSet data) {
        data.shuffle();
        DataNormalization normalizer = new NormalizerStandardize();
        normalizer.fit(data);
        normalizer.transform(data);

        return data.splitTestAndTrain(parameters.getPercentTestTrain());
    }

    /**
     * @param data
     * @param expected
     * @param fold_number
     * @param increase_ratio
     * @return
     */
    public HashMap<MultiLayerNetwork, Double> runKFold(
            final List<List<Double>> data,
            final List<Double> expected,
            final int fold_number,
            final int increase_ratio) {
        TrainingDataset dataset = new TrainingDataset(data, expected);
        List<TrainingDataset> folds = dataset.prepareFolds(fold_number);
        HashMap<MultiLayerNetwork, Double> map = new HashMap<>();
        for (int i = 0; i < fold_number; i++) {
            TrainingDataset testingfold = folds.get(i);
            TrainingDataset learning_fold = new TrainingDataset();
            for (int j = 0; j < fold_number; j++) {
                if (j != i) {
                    learning_fold.addFoldInDataset(folds, i);
                }
            }
            TrainingDataset dataset_increased = learning_fold.increaseTrueAlert(
                    increase_ratio);
            MultiLayerNetwork nn = run(
                    dataset_increased.getData(),
                    dataset_increased.getExpected());
            Double score = modelEvaluation(testingfold, nn);

            map.put(nn, score);
        }
        return  map;
    }


}
