package be.cylab.java.wowa.training;

import java.util.ArrayList;
import java.util.List;

/**
 * Class to manage the training dataset.
 * It uses for the k-fold cross validation to prepare folds.
 */
class TrainingDataset {
    private List<List<Double>> data;
    private List<Double> expected;
    private int length;

    /**
     * Default constructor.
     */
    TrainingDataset() {
        this.data = new ArrayList<>();
        this.expected = new ArrayList<>();
        this.length = 0;
    }

    /**
     * Constructor takes List as arguments.
     *
     * @param data
     * @param expected
     */
    TrainingDataset(
            final List<List<Double>> data,
            final List<Double> expected) {
        if (data.size() != expected.size()) {
            throw new IllegalArgumentException(
                    "Data and expected list must have the same size");
        }
        this.data = data;
        this.expected = expected;
        this.length = data.size();
    }

    /**
     * Getter for data list.
     *
     * @return
     */
    List<List<Double>> getData() {
        return data;
    }

    /**
     * Setter for data list.
     *
     * @param data
     */

    /**
     * Getter for expected list.
     *
     * @return
     */
    public List<Double> getExpected() {
        return expected;
    }


    /**
     * Length getter.
     *
     * @return
     */
    int getLength() {
        return this.length;
    }


    /**
     * Copy an element from a dataset to another one.
     *
     * @param dataset
     * @param index
     * @return
     */
    TrainingDataset addElementInDataset(
            final TrainingDataset dataset,
            final int index) {
        this.data.add(dataset.getData().get(index));
        this.expected.add(dataset.getExpected().get(index));
        this.length++;

        return this;

    }

    /**
     * Remove an element from a dataset.
     *
     * @param index
     */
    void removeElementInDataset(final int index) {
        this.data.remove(index);
        this.expected.remove(index);
        this.length--;
    }

    /**
     * Method to add a fold to a dataset (learning dataset creation).
     *
     * @param data_to_add
     * @param index
     */
    TrainingDataset addFoldInDataset(
            final List<TrainingDataset> data_to_add,
            final int index) {
        this.data.addAll(data_to_add.get(index).data);
        this.expected.addAll(data_to_add.get(index).expected);
        this.length += data_to_add.get(index).getLength();
        return this;
    }

    /**
     * Method to separate randomly the base dataset in X folds dataset.
     *
     * @param fold_number
     * @return
     */
    final List<TrainingDataset> prepareFolds(
            final int fold_number) {
        //List<List<Double>> data = dataset.getData();
        List<Double> expected = this.getExpected();
        List<TrainingDataset> fold_dataset = new ArrayList<>();
        //Check if it is rounded !!!!
        int alert_number
                = (int) Math.floor(Utils.sumListElements(expected)
                / fold_number);
        int no_alert_number = (int) (expected.size()
                - Utils.sumListElements(expected)) / fold_number;

        for (int i = 0; i < fold_number; i++) {
            TrainingDataset tmp = new TrainingDataset();
            int alert_counter = 0;
            int no_alert_counter = 0;
            while (tmp.getLength() < (alert_number + no_alert_number)) {
                int index = Utils.randomInteger(0, this.getLength() - 1);
                if (this.getExpected().get(index) == 1
                        && alert_counter < alert_number) {
                    tmp.addElementInDataset(this, index);
                    this.removeElementInDataset(index);
                    alert_counter++;
                } else if (this.getExpected().get(index) == 0
                        && no_alert_counter < no_alert_number) {
                    tmp.addElementInDataset(this, index);
                    this.removeElementInDataset(index);
                    no_alert_counter++;
                }
            }
            fold_dataset.add(tmp);
        }
        int fold_counter = 0;
        while (this.getLength() > 0) {
            int i = Utils.randomInteger(0, this.getLength() - 1);
            fold_dataset.get(fold_counter).addElementInDataset(this, i);
            this.removeElementInDataset(i);
            if (fold_counter == fold_dataset.size() - 1) {
                fold_counter = 0;
            } else {
                fold_counter++;
            }
        }
        return fold_dataset;
    }

    /**
     * Method to increase the number of true alert in data.
     * to increase penalty to do not detect a true alert
     *
     * @param increase_ratio
     */
    final TrainingDataset increaseTrueAlert(
            final int increase_ratio) {
        int data_size = expected.size();
        for (int i = 0; i < data_size; i++) {
            if (expected.get(i) == 1) {
                for (int j = 0; j < increase_ratio - 1; j++) {
                    expected.add(expected.get(i));
                    data.add(data.get(i));
                }
            }
        }
        return new TrainingDataset(data, expected);
    }
}
